import matplotlib.pyplot as plt
import numpy as np
import networkx as nx
import random


from typing import List


def set_seeds(seed=0):
    random.seed(seed)
    np.random.seed(seed)


class OpinionNetwork(object):
    """Class representing the network of agents with opinion dymanics

    Agents opinionts are represented by an array of shape (N,) with floats.
    Value at i-th index represents i-th agent's opinion.

    Agent network is sampled according to the given network type:
        - albert-barabasi (default)
            network_arg is a list that contains a number of new edges
        - small-world (Watts Strogatz Graph)
            network_arg is a list that contains a number of edges for regular lattice and
            a number representing rewire probability
        - erdos-renyi (Erdos Renyi Graph)
            network_arg is a list containing a number representing the edge creation 
            probability
    """

    def __init__(self, N, d, mu, type='albert-barabasi', ab_arg=[4, 64], sm_arg=[4, 0.3], er_arg=[0.3]) -> None:
        self.N = N
        self.d = d
        self.mu = mu
        self.type = type
        self.opinion_history = []

        # Sampling agent array
        self.agents = np.random.random(size=(N,))

        # Sampling network
        if type == 'albert-barabasi':
            self.g = nx.barabasi_albert_graph(N, ab_arg[0],  ab_arg[1])
        elif type == 'small-world':
            self.g = nx.watts_strogatz_graph(N, sm_arg[0], sm_arg[1])
        elif type == 'erdos-renyi':
            self.g = nx.erdos_renyi_graph(N, er_arg[0])
        else:
            raise ValueError('Type of network graph invalid.\
             Choose either "albert-barabasi", "small-world" or "erdos-renyi".')

    def step(self) -> None:
        """Function for atomic step of the evolution
        - sampling pair of two agents and updating their opinions
        according to the rule:
        o_i(t+1)=o_i(t) + mu(o_j(t)-o_i(t))
        o_j(t+1)=o_j(t) + mu(o_i(t)-o_j(t))
        """
        # Pick pair of agents
        agent_i = np.random.randint(0, self.N)
        agent_j = np.random.choice(list(self.g.adj[agent_i]))
        # Extract agents opinions
        o_i = self.agents[agent_i]
        o_j = self.agents[agent_j]
        # Check if opinions are close
        if np.abs(o_i-o_j) < self.d:
            # Perform opinion update
            new_o_i = o_i + self.mu*(o_j-o_i)
            new_o_j = o_j + self.mu*(o_i-o_j)
            agents_copy = self.agents.copy()
            self.agents[agent_i] = new_o_i
            self.agents[agent_j] = new_o_j
        else:
            pass

    def run_evolution(self, T) -> List:
        """Run evolution of the dynamics and save distribution of opinions
        at each time step.

        Return array with histories.
        """
        # Initalize history
        self.opinion_history = [self.agents.copy()]
        # For each time step
        for i in range(T):
            # Perform update for N//2 pairs of agents
            for j in range(self.N//2):
                self.step()
            # Append opinion distribution to the history
            self.opinion_history.append(self.agents.copy())
        return self.opinion_history

    def plot_evolution(self, times=None, suffix=''):
        """Function to plot evolution of opinions.

        The function plots the opinion distribution at 
        given indices of history (times list).
        By default all.
        """
        history = self.opinion_history

        if times is None:
            times = list(range(len(history)))

        plt.figure()
        for index in times:
            time_index_ar = np.ones_like(history[index])*index
            plt.scatter(time_index_ar,
                        history[index], color='tab:blue', marker='.')
        plt.xlabel('time')
        plt.ylabel('opinions')
        plt.title(self.type + suffix)
        plt.show()

    def run_one_timestep(self):
        for j in range(self.N//2):
            self.step()
        self.opinion_history.append(self.agents.copy())


class Market(object):
    """Class representing the market.

    Arguments:
        - n - OpinionNetwork object with agents and their opinions
        - pi_0 - initial price
        - p - probability of participation in market (or proportionality constant)
        - gamma - proportionality constant for excessive demand

    """

    def __init__(self, n: OpinionNetwork, pi_0=0.5, p=0.2, gamma=0.1) -> None:
        self.n = n
        self.N = n.N
        self.p = p
        self.gamma = gamma
        self.price_history = [pi_0]
        # history of demand
        self.demand_history = []
        # history of number of participating agents
        self.participating_agents_history = []
        # volume history
        self.volume_history = []

    def run_one_timestep(self, agent_ids) -> None:
        """Function for running one step of the market.

        Arguments:
            - agent_ids - ids of agents possibly participating in the market
        """
        # Initialize variables
        demand = 0
        current_price = self.price_history[-1]
        participating_agents = 0
        volume = 0
        # Iterate over possible participating agents
        for agent_id in agent_ids:
            # Get opinion of current agent
            opinion = self.n.agents[agent_id]
            # Find agent's demand
            agent_demand = self.gamma*(opinion-current_price)
            # Check if agent buys
            if self.n.agents[agent_id] > current_price:
                # Increment number of participating agents in the time step
                participating_agents += 1
                # Increase volume in the time step
                volume += agent_demand
            # Aggregate demand irrespective of whether agent bought
            demand += agent_demand
        # Update histories
        self.demand_history.append(demand)
        self.volume_history.append(volume)
        self.participating_agents_history.append(participating_agents)

        # Update price
        price_change = demand*np.abs(np.random.normal(0, 0.05))
        # price_change = demand*(np.random.normal(0,0.05))
        new_price = current_price + np.round(price_change, 2)
        if new_price <= 0.01:
            new_price = 0.01
        if new_price >= 0.99:
            new_price = 0.99
        self.price_history.append(new_price)

    def pick_agents_const_prob(self, p):
        """ Function for picking agents according to binomial distribution

        Arguments:
            - probability of pick for each agent

        """
        agent_ids = []
        for idx in range(self.N):
            if np.random.random() < self.p:
                agent_ids.append(idx)
        return agent_ids

    def run_evolution(self, T, picking_type='const'):
        """Function for running evolution of the market

        Arguments:
            - T - number of time steps
            - picking_type - method of picking agents (default: 'const')
                if 'const' probability of pick will be constant with time
                if 'power_law' robability will follow power law

        """
        lambda_coef = 2.44  # Coefficient for power law
        for t in range(T):
            # Run one timestep of opinion dynamics network
            self.n.run_one_timestep()
            # Pick agents either according to constant w.r.t. time or power law probability
            if picking_type == 'const':
                agent_ids = self.pick_agents_const_prob(self.p)
            elif picking_type == 'power_law':
                # prob = (T-t+1)**(-lambda_coef)+self.p
                # print(prob)
                agent_ids = self.pick_agents_const_prob(
                    (T-t+1)**(-lambda_coef)+0.01)
            else:
                raise ValueError('Invalid picking method')
            # Run one timestep of the market
            self.run_one_timestep(agent_ids)
